@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_proto_modules.c,v 1.6.2.1 2009/04/08 11:11:29 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_proto_modules.c,v 1.6.2.9 2009/08/22 04:19:49 t-ishii Exp $
  * 
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -46,6 +46,7 @@
 #include "pool.h"
 #include "pool_signal.h"
 #include "pool_proto_modules.h"
+#include "parser/pool_string.h"
 
 int force_replication;
 int replication_was_enabled;		/* replication mode was enabled */
@@ -87,7 +88,10 @@ char query_string_buffer[QUERY_STRING_BUFFER_LEN];
  */
 char *parsed_query = NULL;
 
-POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend, 
+static int check_errors(POOL_CONNECTION_POOL *backend, int backend_id);
+static void generate_error_message(char *prefix, int specific_error, char *query);
+
+POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 										POOL_CONNECTION_POOL *backend)
 {
 	int pid, pid1;
@@ -127,7 +131,7 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
  * Process Query('Q') message
  * Query messages include a SQL string.
  */
- POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend, 
+ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 						 POOL_CONNECTION_POOL *backend, char *query)
 {
 	char *string, *string1;
@@ -137,9 +141,7 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 	List *parse_tree_list;
 	Node *node = NULL, *node1;
 	POOL_STATUS status;
-	int serialization_error_detected = 0;
-	int deadlock_detected = 0;
-	int	active_sql_transaction_error = 0;
+	int specific_error;
 
 	POOL_MEMORY_POOL *old_context = NULL;
 	Portal *portal;
@@ -229,19 +231,19 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 
 		if (pool_config->parallel_mode)
 		{
-      /* The Query is analyzed first in a parallel mode(in_parallel_query), 
+      /* The Query is analyzed first in a parallel mode(in_parallel_query),
        * and, next, the Query is rewritten(rewrite_query_stmt).
        */
- 
+
 			/* analyze the query */
 			RewriteQuery *r_query = is_parallel_query(node,backend);
 
 			if(r_query->is_loadbalance)
 			{
-        /* Usual processing of pgpool is done by using the rewritten Query 
-         * if judged a possible load-balancing as a result of analyzing 
-         * the Query. 
-         * Of course, the load is distributed only for load_balance_mode=true. 
+        /* Usual processing of pgpool is done by using the rewritten Query
+         * if judged a possible load-balancing as a result of analyzing
+         * the Query.
+         * Of course, the load is distributed only for load_balance_mode=true.
          */
 				if(r_query->r_code ==  SEND_LOADBALANCE_ENGINE)
 				{
@@ -253,8 +255,8 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 				pool_debug("SimpleQuery: loadbalance_query =%s",string);
 			}
 			else if (r_query->is_parallel)
-			{ 
-				/* 
+			{
+				/*
 				 * For the Query that the parallel processing is possible.
 				 * Call parallel exe engine and return status to the upper layer.
 				 */
@@ -504,7 +506,7 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 			master_slave_was_enabled = 1;
 			MASTER_SLAVE = 0;
 			master_slave_dml = 1;
-		}		
+		}
 	}
 
 	if (MAJOR(backend) == PROTO_MAJOR_V2 && is_start_transaction_query(node))
@@ -514,81 +516,44 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 
 	if (REPLICATION || PARALLEL_MODE)
 	{
-		/* check if query is "COMMIT" */
+		/* check if query is "COMMIT" or "ROLLBACK" */
 		commit = is_commit_query(node);
 		free_parser();
 
-		/* send query to master node */
+		/*
+		 * Query is not commit/rollback
+		 */
 		if (!commit)
 		{
+			/* Send the query to master node */
+
 			if (send_simplequery_message(MASTER(backend), len, string, MAJOR(backend)) != POOL_CONTINUE)
 				return POOL_END;
 
-			if (wait_for_query_response(MASTER(backend), string) != POOL_CONTINUE)
-				return POOL_END;
+			if (wait_for_query_response(frontend, MASTER(backend), string, MAJOR(backend)) != POOL_CONTINUE)
+			{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
 
-			/*
-			 * Check dead lock error on the master node and abort
-			 * transactions on all nodes if so.
-			 */
-			deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
-			if (deadlock_detected < 0)
-				return POOL_END;
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
 
-			/*
-			 * Check serialization failure error and abort
-			 * transactions on all nodes if so. Otherwise we allow
-			 * data inconsistency among DB nodes. See following
-			 * scenario: (M:master, S:slave)
-			 *
-			 * M:S1:BEGIN;
-			 * M:S2:BEGIN;
-			 * S:S1:BEGIN;
-			 * S:S2:BEGIN;
-			 * M:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * M:S2:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * S:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * S:S2:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * M:S1:UPDATE t1 SET i = i + 1;
-			 * S:S1:UPDATE t1 SET i = i + 1;
-			 * M:S2:UPDATE t1 SET i = i + 1; <-- blocked
-			 * S:S1:COMMIT;
-			 * M:S1:COMMIT;
-			 * M:S2:ERROR:  could not serialize access due to concurrent update
-			 * S:S2:UPDATE t1 SET i = i + 1; <-- success in UPDATE and data becomes inconsistent!
-			 */
-			serialization_error_detected = detect_serialization_error(MASTER(backend), MAJOR(backend));
-			if (serialization_error_detected < 0)
 				return POOL_END;
+			}
 
-			/*
-			 * check "SET TRANSACTION ISOLATION LEVEL must be called before any query" error.
-			 * This happens in following scenario:
-			 * 
-			 * M:S1:BEGIN;
-			 * S:S1:BEGIN;
-			 * M:S1:SELECT 1; <-- only sent to MASTER
-			 * M:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * S:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * M: <-- error
-			 * S: <-- ok since no previous SELECT is sent. kind mismatch error occurs!
-			 */
-			active_sql_transaction_error = detect_active_sql_transaction_error(MASTER(backend), MAJOR(backend));
-			if (active_sql_transaction_error < 0)
-				return POOL_END;
-		}
+			/* Check specific errors */
+			specific_error = check_errors(backend, MASTER_NODE_ID);
+			if (specific_error)
+			{
+				/* log error message */
+				generate_error_message("SimpleQuery: ", specific_error, string);
 
-		if (deadlock_detected == SPECIFIED_ERROR || serialization_error_detected == SPECIFIED_ERROR ||
-			active_sql_transaction_error == SPECIFIED_ERROR)
-		{
-			if (deadlock_detected == SPECIFIED_ERROR)
-				pool_log("SimpleQuery: received deadlock error message from master node. query: %s", string);
-			else if (serialization_error_detected == SPECIFIED_ERROR)
-				pool_log("SimpleQuery: received serialization failure error message from master node. query: %s", string);
-			else
-				pool_log("SimpleQuery: received SET TRANSACTION ISOLATION LEVEL must be called before any query error. query: %s", string);
-			string = POOL_ERROR_QUERY;
-			len = strlen(string) + 1;
+				/* Set error query to abort transactions on other nodes */
+				string = POOL_ERROR_QUERY;
+				len = strlen(string) + 1;
+			}
 		}
 
 		/* send query to other than master nodes */
@@ -601,24 +566,45 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 				return POOL_END;
 		}
 
-		/* wait for response except MASTER node */
+		/* Wait for nodes othan than the master node */
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
 			if (!VALID_BACKEND(i) || IS_MASTER_NODE_ID(i))
 				continue;
 
-			if (wait_for_query_response(CONNECTION(backend, i), string) != POOL_CONTINUE)
+			if (wait_for_query_response(frontend, CONNECTION(backend, i), string, MAJOR(backend)) != POOL_CONTINUE)
+			{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
+
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
+
 				return POOL_END;
+			}
 		}
 
-		/* send "COMMIT" to master node if query is "COMMIT" */
+		/* send "COMMIT" or "ROLLBACK" to only master node if query is "COMMIT" or "ROLLBACK" */
 		if (commit)
 		{
 			if (send_simplequery_message(MASTER(backend), len, string, MAJOR(backend)) != POOL_CONTINUE)
 				return POOL_END;
 
-			if (wait_for_query_response(MASTER(backend), string) != POOL_CONTINUE)
+			if (wait_for_query_response(frontend, MASTER(backend), string, MAJOR(backend)) != POOL_CONTINUE)
+			{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
+
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
+
 				return POOL_END;
+			}
+
 
 			TSTATE(backend) = 'I';
 		}
@@ -629,8 +615,18 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 		if (send_simplequery_message(MASTER(backend), len, string, MAJOR(backend)) != POOL_CONTINUE)
 			return POOL_END;
 
-		if (wait_for_query_response(MASTER(backend), string) != POOL_CONTINUE)
-			return POOL_END;
+		if (wait_for_query_response(frontend, MASTER(backend), string, MAJOR(backend)) != POOL_CONTINUE)
+		{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
+
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
+
+				return POOL_END;
+		}
 	}
 
 	return POOL_CONTINUE;
@@ -639,7 +635,7 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 /*
  * process EXECUTE (V3 only)
  */
-POOL_STATUS Execute(POOL_CONNECTION *frontend, 
+POOL_STATUS Execute(POOL_CONNECTION *frontend,
 						   POOL_CONNECTION_POOL *backend)
 {
 	char *string;		/* portal name + null terminate + max_tobe_returned_rows */
@@ -650,10 +646,8 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 	Portal *portal;
 	char *string1;
 	PrepareStmt *p_stmt;
-	int deadlock_detected = 0;
-	int serialization_error_detected = 0;
-	int	active_sql_transaction_error = 0;
 	POOL_STATUS ret;
+	int specific_error = 0;
 
 	/* read Execute packet */
 	if (pool_read(frontend, &len, sizeof(len)) < 0)
@@ -676,6 +670,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 
 		string1 = portal->sql_string;
 		node = (Node *)p_stmt->query;
+		strncpy(query_string_buffer, string1, sizeof(query_string_buffer));
 
 		if ((IsA(node, PrepareStmt) || IsA(node, DeallocateStmt) ||
 			 IsA(node, VariableSetStmt)) &&
@@ -721,6 +716,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 			return POOL_END;
 		}
 */
+		/* check if query is "COMMIT" or "ROLLBACK" */
 		commit = is_commit_query((Node *)p_stmt->query);
 	}
 
@@ -738,49 +734,37 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 
 	if (REPLICATION || PARALLEL_MODE)
 	{
-		/* send query to master node */
+		/*
+		 * Query is not commit/rollback
+		 */
 		if (!commit)
 		{
+			/* Send the query to master node */
+
 			if (send_execute_message(backend, MASTER_NODE_ID, len, string) != POOL_CONTINUE)
 				return POOL_END;
 
-			pool_debug("waiting for backend completing the query");
-			if (synchronize(CONNECTION(backend, MASTER_NODE_ID)))
-				return POOL_END;
+			if (wait_for_query_response(frontend, MASTER(backend), string, MAJOR(backend)) != POOL_CONTINUE)
+			{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
 
-			/*
-			 * Check dead lock error on the master node and abort
-			 * transactions on all nodes if so.
-			 */
-			deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
-			if (deadlock_detected < 0)
-				return POOL_END;
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
 
-			/*
-			 * Check serialization failure error and abort all nodes
-			 * if so. Otherwise we allow data inconsistency among DB
-			 * nodes. See following scenario: (M:master, S:slave)
-			 */
-			serialization_error_detected = detect_serialization_error(MASTER(backend), MAJOR(backend));
-			if (serialization_error_detected < 0)
 				return POOL_END;
+			}
 
-			/*
-			 * check "SET TRANSACTION ISOLATION LEVEL must be called before any query" error.
-			 * This happens in following scenario:
-			 * 
-			 * M:S1:BEGIN;
-			 * S:S1:BEGIN;
-			 * M:S1:SELECT 1; <-- only sent to MASTER
-			 * M:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * S:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
-			 * M: <-- error
-			 * S: <-- ok since no previous SELECT is sent. kind mismatch error occurs!
-			 */
-			active_sql_transaction_error = detect_active_sql_transaction_error(MASTER(backend), MAJOR(backend));
-			if (active_sql_transaction_error < 0)
-				return POOL_END;
 
+			/* Check specific errors */
+			specific_error = check_errors(backend, MASTER_NODE_ID);
+			if (specific_error)
+			{
+				/* log error message */
+				generate_error_message("Execute: ", specific_error, string);
+			}
 		}
 
 		/* send query to other nodes */
@@ -788,19 +772,12 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 		{
 			if (!VALID_BACKEND(i) || IS_MASTER_NODE_ID(i))
 				continue;
-			if (deadlock_detected == SPECIFIED_ERROR || serialization_error_detected == SPECIFIED_ERROR ||
-				active_sql_transaction_error == SPECIFIED_ERROR)
+
+			if (specific_error)
 			{
 				char msg[1024] = "pgpoool_error_portal"; /* large enough */
 				int len = strlen(msg);
 
-				if (deadlock_detected == SPECIFIED_ERROR)
-					pool_log("Execute: received deadlock error message from master node. query: %s", string);
-				else if (serialization_error_detected == SPECIFIED_ERROR)
-					pool_log("SimpleQuery: received serialization failure error message from master node. query: %s", string);
-				else
-					pool_log("SimpleQuery: received SET TRANSACTION ISOLATION LEVEL must be called before any query error. query: %s", string);
-
 				memset(msg + len, 0, sizeof(int));
 				if (send_execute_message(backend, i, len + 5, msg))
 					return POOL_END;
@@ -809,25 +786,44 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 				return POOL_END;
 		}
 
-		/* wait for nodes excepted for master node */
+		/* Wait for nodes other than the master node */
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
 			if (!VALID_BACKEND(i) || IS_MASTER_NODE_ID(i))
 				continue;
 
-			pool_debug("waiting for backend completing the query");
-			if (synchronize(CONNECTION(backend, i)))
+			if (wait_for_query_response(frontend, CONNECTION(backend, i), string, MAJOR(backend)) != POOL_CONTINUE)
+			{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
+
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
+
 				return POOL_END;
+			}
 		}
 
+		/* send "COMMIT" or "ROLLBACK" to only master node if query is "COMMIT" or "ROLLBACK" */
 		if (commit)
 		{
 			if (send_execute_message(backend, MASTER_NODE_ID, len, string) != POOL_CONTINUE)
 				return POOL_END;
 
-			pool_debug("waiting for backend completing the query");
-			if (synchronize(MASTER(backend)))
+			if (wait_for_query_response(frontend, MASTER(backend), string, MAJOR(backend)) != POOL_CONTINUE)
+			{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
+
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
+
 				return POOL_END;
+			}
 		}
 	}
 	else
@@ -835,9 +831,18 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 		if (send_execute_message(backend, MASTER_NODE_ID, len, string) != POOL_CONTINUE)
 			return POOL_END;
 
-		pool_debug("waiting for backend completing the query");
-		if (synchronize(CONNECTION(backend, MASTER_NODE_ID)))
-			return POOL_END;
+		if (wait_for_query_response(frontend, MASTER(backend), string, MAJOR(backend)) != POOL_CONTINUE)
+		{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
+
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
+
+				return POOL_END;
+		}
 	}
 
 	while ((ret = read_kind_from_backend(frontend, backend, &kind)) == POOL_CONTINUE)
@@ -866,7 +871,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 /*
  * process Parse (V3 only)
  */
-POOL_STATUS Parse(POOL_CONNECTION *frontend, 
+POOL_STATUS Parse(POOL_CONNECTION *frontend,
 						 POOL_CONNECTION_POOL *backend)
 {
 	char kind;
@@ -1031,7 +1036,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend,
 			if (!VALID_BACKEND(i) || IS_MASTER_NODE_ID(i))
 				continue;
 
-			pool_debug("waiting for %th backend completing the query", i);
+			pool_debug("waiting for %dth backend completing the query", i);
 			if (synchronize(CONNECTION(backend, i)))
 				return POOL_END;
 		}
@@ -1044,7 +1049,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend,
 
 		if (ret != POOL_CONTINUE)
 			return ret;
-			
+
 		SimpleForwardToFrontend(kind, frontend, backend);
 		if (pool_flush(frontend) < 0)
 			return POOL_ERROR;
@@ -1063,7 +1068,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend,
  *	 to all DB nodes to abort transaction.
  * - internal transaction is closed
  */
-POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend, 
+POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 								 POOL_CONNECTION_POOL *backend, int send_ready)
 {
 	StartupPacket *sp;
@@ -1132,7 +1137,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		mismatch_ntuples = 0;
 	}
 
-	/* 
+	/*
 	 * if a transaction is started for insert lock, we need to close
 	 * the transaction.
 	 */
@@ -1186,7 +1191,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 			{
 				if (!VALID_BACKEND(i) || IS_MASTER_NODE_ID(i))
 					continue;
-				
+
 				if (pool_read(CONNECTION(backend, i), &kind1, sizeof(kind)))
 					return POOL_END;
 			}
@@ -1251,10 +1256,10 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 
 	sp = MASTER_CONNECTION(backend)->sp;
 	if (MASTER(backend)->tstate == 'T')
-		snprintf(psbuf, sizeof(psbuf), "%s %s %s idle in transaction", 
+		snprintf(psbuf, sizeof(psbuf), "%s %s %s idle in transaction",
 				 sp->user, sp->database, remote_ps_data);
 	else
-		snprintf(psbuf, sizeof(psbuf), "%s %s %s idle", 
+		snprintf(psbuf, sizeof(psbuf), "%s %s %s idle",
 				 sp->user, sp->database, remote_ps_data);
 	set_ps_display(psbuf, false);
 
@@ -1262,7 +1267,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 }
 
 
-POOL_STATUS FunctionCall(POOL_CONNECTION *frontend, 
+POOL_STATUS FunctionCall(POOL_CONNECTION *frontend,
 								POOL_CONNECTION_POOL *backend)
 {
 	char dummy[2];
@@ -1359,7 +1364,7 @@ POOL_STATUS FunctionCall(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS FunctionResultResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS FunctionResultResponse(POOL_CONNECTION *frontend,
 										  POOL_CONNECTION_POOL *backend)
 {
 	char dummy;
@@ -1421,7 +1426,7 @@ POOL_STATUS FunctionResultResponse(POOL_CONNECTION *frontend,
 	return pool_flush(frontend);
 }
 
-POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 										   POOL_CONNECTION_POOL *backend)
 {
 	char fkind;
@@ -1442,6 +1447,8 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 
 	if (receive_extended_begin)
 	{
+		receive_extended_begin = 0;
+
 		/* send sync message */
 		send_extended_protocol_message(backend, MASTER_NODE_ID, "S", 0, "");
 
@@ -1520,7 +1527,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	return status;
 }
 
-POOL_STATUS CompleteCommandResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS CompleteCommandResponse(POOL_CONNECTION *frontend,
 										   POOL_CONNECTION_POOL *backend)
 {
 	int i;
@@ -1548,8 +1555,8 @@ POOL_STATUS CompleteCommandResponse(POOL_CONNECTION *frontend,
 		if (len != len1)
 		{
 			pool_debug("Complete Command Response: message length does not match between master(%d \"%s\",) and %d th server (%d \"%s\",)",
-					   len, string, len1, string1);
-			
+					   len, string, i, len1, string1);
+
 			free(string1);
 			return POOL_END;
 		}
@@ -1567,7 +1574,7 @@ POOL_STATUS CompleteCommandResponse(POOL_CONNECTION *frontend,
 	return pool_flush(frontend);
 }
 
-int RowDescription(POOL_CONNECTION *frontend, 
+int RowDescription(POOL_CONNECTION *frontend,
 						  POOL_CONNECTION_POOL *backend,
 						  short *result)
 {
@@ -1623,7 +1630,7 @@ int RowDescription(POOL_CONNECTION *frontend,
 				if (len != len1)
 				{
 					pool_error("RowDescription: field length does not match between backends master(%d) and %d th backend(%d)",
-							   ntohl(len), ntohl(len1));
+							   ntohl(len), j, ntohl(len1));
 					return POOL_FATAL;
 				}
 			}
@@ -1694,7 +1701,7 @@ int RowDescription(POOL_CONNECTION *frontend,
 	return pool_flush(frontend);
 }
 
-POOL_STATUS AsciiRow(POOL_CONNECTION *frontend, 
+POOL_STATUS AsciiRow(POOL_CONNECTION *frontend,
 							POOL_CONNECTION_POOL *backend,
 							short num_fields)
 {
@@ -1806,7 +1813,7 @@ POOL_STATUS AsciiRow(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS BinaryRow(POOL_CONNECTION *frontend, 
+POOL_STATUS BinaryRow(POOL_CONNECTION *frontend,
 							 POOL_CONNECTION_POOL *backend,
 							 short num_fields)
 {
@@ -1858,14 +1865,14 @@ POOL_STATUS BinaryRow(POOL_CONNECTION *frontend,
 		{
 			/* field size */
 			if (pool_read(MASTER(backend), &size, sizeof(int)) < 0)
-				return POOL_END;			
+				return POOL_END;
 			for (j=0;j<NUM_BACKENDS;j++)
 			{
 				if (VALID_BACKEND(j) && !IS_MASTER_NODE_ID(j))
 				{
 					/* field size */
 					if (pool_read(CONNECTION(backend, i), &size, sizeof(int)) < 0)
-						return POOL_END;			
+						return POOL_END;
 
 					/* XXX: field size maybe different among
 					   backends. If we were a paranoid, we have to treat
@@ -1908,7 +1915,7 @@ POOL_STATUS BinaryRow(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS CursorResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS CursorResponse(POOL_CONNECTION *frontend,
 								  POOL_CONNECTION_POOL *backend)
 {
 	char *string = NULL;
@@ -1935,7 +1942,7 @@ POOL_STATUS CursorResponse(POOL_CONNECTION *frontend,
 			{
 				pool_error("CursorResponse: length does not match between master(%d) and %d th backend(%d)",
 						   len, i, len1);
-				pool_error("CursorResponse: master(%s) %d th backend(%s)", string1, string);
+				pool_error("CursorResponse: master(%s) %d th backend(%s)", string1, i, string);
 				free(string1);
 				return POOL_END;
 			}
@@ -1957,7 +1964,7 @@ POOL_STATUS CursorResponse(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS ErrorResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS ErrorResponse(POOL_CONNECTION *frontend,
 						  POOL_CONNECTION_POOL *backend)
 {
 	char *string = NULL;
@@ -1985,11 +1992,11 @@ POOL_STATUS ErrorResponse(POOL_CONNECTION *frontend,
 		TSTATE(backend) = 'E';
 	else
 		TSTATE(backend) = 'I';
-			
+
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS NoticeResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS NoticeResponse(POOL_CONNECTION *frontend,
 								  POOL_CONNECTION_POOL *backend)
 {
 	char *string = NULL;
@@ -2016,7 +2023,7 @@ POOL_STATUS NoticeResponse(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS CopyInResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS CopyInResponse(POOL_CONNECTION *frontend,
 								  POOL_CONNECTION_POOL *backend)
 {
 	POOL_STATUS status;
@@ -2037,7 +2044,7 @@ POOL_STATUS CopyInResponse(POOL_CONNECTION *frontend,
 	return status;
 }
 
-POOL_STATUS CopyOutResponse(POOL_CONNECTION *frontend, 
+POOL_STATUS CopyOutResponse(POOL_CONNECTION *frontend,
 								   POOL_CONNECTION_POOL *backend)
 {
 	POOL_STATUS status;
@@ -2227,7 +2234,7 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 			}
 		}
 		else
-			pool_write(frontend, string, len);		
+			pool_write(frontend, string, len);
 
 		if (len == PROTO_MAJOR_V3)
 		{
@@ -2280,3 +2287,90 @@ POOL_STATUS EmptyQueryResponse(POOL_CONNECTION *frontend,
 	pool_write(frontend, "I", 1);
 	return pool_write_and_flush(frontend, "", 1);
 }
+
+/*
+ * Check various errors from backend.  return values: 0: no error 1:
+ * deadlock detected 2: serialization error detected 3: query cancel
+ * detected: 4
+ */
+static int check_errors(POOL_CONNECTION_POOL *backend, int backend_id)
+{
+
+	/*
+	 * Check dead lock error on the master node and abort
+	 * transactions on all nodes if so.
+	 */
+	if (detect_deadlock_error(CONNECTION(backend, backend_id), MAJOR(backend)) == SPECIFIED_ERROR)
+		return 1;
+
+	/*
+	 * Check serialization failure error and abort
+	 * transactions on all nodes if so. Otherwise we allow
+	 * data inconsistency among DB nodes. See following
+	 * scenario: (M:master, S:slave)
+	 *
+	 * M:S1:BEGIN;
+	 * M:S2:BEGIN;
+	 * S:S1:BEGIN;
+	 * S:S2:BEGIN;
+	 * M:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
+	 * M:S2:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
+	 * S:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
+	 * S:S2:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
+	 * M:S1:UPDATE t1 SET i = i + 1;
+	 * S:S1:UPDATE t1 SET i = i + 1;
+	 * M:S2:UPDATE t1 SET i = i + 1; <-- blocked
+	 * S:S1:COMMIT;
+	 * M:S1:COMMIT;
+	 * M:S2:ERROR:  could not serialize access due to concurrent update
+	 * S:S2:UPDATE t1 SET i = i + 1; <-- success in UPDATE and data becomes inconsistent!
+	 */
+	if (detect_serialization_error(CONNECTION(backend, backend_id), MAJOR(backend)) == SPECIFIED_ERROR)
+		return 2;
+
+	/*
+	 * check "SET TRANSACTION ISOLATION LEVEL must be called before any query" error.
+	 * This happens in following scenario:
+	 *
+	 * M:S1:BEGIN;
+	 * S:S1:BEGIN;
+	 * M:S1:SELECT 1; <-- only sent to MASTER
+	 * M:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
+	 * S:S1:SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;
+	 * M: <-- error
+	 * S: <-- ok since no previous SELECT is sent. kind mismatch error occurs!
+	 */
+	if (detect_active_sql_transaction_error(CONNECTION(backend, backend_id), MAJOR(backend)) == SPECIFIED_ERROR)
+		return 3;
+
+	/* check query cancel error */
+	if (detect_query_cancel_error(CONNECTION(backend, backend_id), MAJOR(backend)) == SPECIFIED_ERROR)
+		return 4;
+
+	return 0;
+}
+
+static void generate_error_message(char *prefix, int specific_error, char *query)
+{
+	static char *error_messages[] = {
+		"received deadlock error message from master node. query: %s",
+		"received serialization failure error message from master node. query: %s",
+		"received SET TRANSACTION ISOLATION LEVEL must be called before any query error. query: %s",
+		"received query cancel error message from master node. query: %s"
+	};
+
+	String *msg;
+
+	if (specific_error < 1 || specific_error > sizeof(error_messages)/sizeof(char *))
+	{
+		pool_error("generate_error_message: invalid specific_error: %d", specific_error);
+		return;
+	}
+
+	specific_error--;
+
+	msg = init_string(prefix);
+	string_append_char(msg, error_messages[specific_error]);
+	pool_error(msg->data, query);
+	free_string(msg);
+}
